@@ -39,8 +39,8 @@ module Control.Foldl (
     , scan
 
     -- * Folds
-    , mconcat
-    , foldMap
+    , Control.Foldl.mconcat
+    , Control.Foldl.foldMap
     , head
     , last
     , lastDef
@@ -60,6 +60,7 @@ module Control.Foldl (
     , index
     , elemIndex
     , findIndex
+    , random
 
     -- * Generic Folds
     , genericLength
@@ -67,6 +68,7 @@ module Control.Foldl (
 
     -- * Container folds
     , list
+    , revList
     , nub
     , eqNub
     , set
@@ -78,10 +80,14 @@ module Control.Foldl (
     , impurely
     , generalize
     , simplify
+    , _Fold1
     , premap
     , premapM
-    , pretraverse
-    , pretraverseM
+    , Handler
+    , handles
+    , EndoM(..)
+    , HandlerM
+    , handlesM
 
     -- * Re-exports
     -- $reexports
@@ -92,18 +98,20 @@ module Control.Foldl (
 
 import Control.Applicative (Applicative(pure, (<*>)),liftA2)
 import Control.Foldl.Internal (Maybe'(..), lazy, Either'(..), hush)
-import Control.Monad ((<=<))
+import Control.Monad ((>=>))
 import Control.Monad.Primitive (PrimMonad)
 import Data.Foldable (Foldable)
 import qualified Data.Foldable as F
 import Data.Functor.Constant (Constant(Constant, getConstant))
 import Data.Functor.Identity (Identity, runIdentity)
+import Data.Profunctor
 import Data.Monoid (Monoid(mempty, mappend), Endo(Endo, appEndo))
 import Data.Vector.Generic (Vector)
 import qualified Data.Vector.Generic as V
 import qualified Data.Vector.Generic.Mutable as M
 import qualified Data.List as List
 import qualified Data.Set as Set
+import System.Random.MWC (createSystemRandom, uniformR)
 import Prelude hiding
     ( head
     , last
@@ -126,8 +134,13 @@ import Prelude hiding
 
     This allows the 'Applicative' instance to assemble derived folds that
     traverse the container only once
+
+    A \''Fold' a b\' processes elements of type __a__ and results in a
+    value of type __b__.
 -}
-data Fold a b = forall x . Fold (x -> a -> x) x (x -> b)
+data Fold a b
+  -- | @Fold @ @ step @ @ initial @ @ extract@
+  = forall x. Fold (x -> a -> x) x (x -> b)
 
 data Pair a b = Pair !a !b
 
@@ -135,24 +148,123 @@ instance Functor (Fold a) where
     fmap f (Fold step begin done) = Fold step begin (f . done)
     {-# INLINABLE fmap #-}
 
+instance Profunctor Fold where
+    lmap = premap
+    rmap = fmap
+
 instance Applicative (Fold a) where
     pure b    = Fold (\() _ -> ()) () (\() -> b)
     {-# INLINABLE pure #-}
+
     (Fold stepL beginL doneL) <*> (Fold stepR beginR doneR) =
         let step (Pair xL xR) a = Pair (stepL xL a) (stepR xR a)
             begin = Pair beginL beginR
-            done (Pair xL xR) = (doneL xL) (doneR xR)
+            done (Pair xL xR) = doneL xL (doneR xR)
         in  Fold step begin done
     {-# INLINABLE (<*>) #-}
 
 instance Monoid b => Monoid (Fold a b) where
     mempty = pure mempty
     {-# INLINABLE mempty #-}
+
     mappend = liftA2 mappend
     {-# INLINABLE mappend #-}
 
--- | Like 'Fold', but monadic
-data FoldM m a b = forall x . FoldM (x -> a -> m x) (m x) (x -> m b)
+instance Num b => Num (Fold a b) where
+    fromInteger = pure . fromInteger
+    {-# INLINABLE fromInteger #-}
+
+    negate = fmap negate
+    {-# INLINABLE negate #-}
+
+    abs = fmap abs
+    {-# INLINABLE abs #-}
+
+    signum = fmap signum
+    {-# INLINABLE signum #-}
+
+    (+) = liftA2 (+)
+    {-# INLINABLE (+) #-}
+
+    (*) = liftA2 (*)
+    {-# INLINABLE (*) #-}
+
+    (-) = liftA2 (-)
+    {-# INLINABLE (-) #-}
+
+instance Fractional b => Fractional (Fold a b) where
+    fromRational = pure . fromRational
+    {-# INLINABLE fromRational #-}
+
+    recip = fmap recip
+    {-# INLINABLE recip #-}
+
+    (/) = liftA2 (/)
+    {-# INLINABLE (/) #-}
+
+instance Floating b => Floating (Fold a b) where
+    pi = pure pi
+    {-# INLINABLE pi #-}
+
+    exp = fmap exp
+    {-# INLINABLE exp #-}
+
+    sqrt = fmap sqrt
+    {-# INLINABLE sqrt #-}
+
+    log = fmap log
+    {-# INLINABLE log #-}
+
+    sin = fmap sin
+    {-# INLINABLE sin #-}
+
+    tan = fmap tan
+    {-# INLINABLE tan #-}
+
+    cos = fmap cos
+    {-# INLINABLE cos #-}
+
+    asin = fmap sin
+    {-# INLINABLE asin #-}
+
+    atan = fmap atan
+    {-# INLINABLE atan #-}
+
+    acos = fmap acos
+    {-# INLINABLE acos #-}
+
+    sinh = fmap sinh
+    {-# INLINABLE sinh #-}
+
+    tanh = fmap tanh
+    {-# INLINABLE tanh #-}
+
+    cosh = fmap cosh
+    {-# INLINABLE cosh #-}
+
+    asinh = fmap asinh
+    {-# INLINABLE asinh #-}
+
+    atanh = fmap atanh
+    {-# INLINABLE atanh #-}
+
+    acosh = fmap acosh
+    {-# INLINABLE acosh #-}
+
+    (**) = liftA2 (**)
+    {-# INLINABLE (**) #-}
+
+    logBase = liftA2 logBase
+    {-# INLINABLE logBase #-}
+
+{-| Like 'Fold', but monadic.
+
+    A \''FoldM' m a b\' processes elements of type __a__ and
+    results in a monadic value of type __m b__.
+-}
+data FoldM m a b =
+  -- | @FoldM @ @ step @ @ initial @ @ extract@
+  forall x . FoldM (x -> a -> m x) (m x) (x -> m b)
 
 instance Monad m => Functor (FoldM m a) where
     fmap f (FoldM step start done) = FoldM step start done'
@@ -165,6 +277,7 @@ instance Monad m => Functor (FoldM m a) where
 instance Monad m => Applicative (FoldM m a) where
     pure b = FoldM (\() _ -> return ()) (return ()) (\() -> return b)
     {-# INLINABLE pure #-}
+
     (FoldM stepL beginL doneL) <*> (FoldM stepR beginR doneR) =
         let step (Pair xL xR) a = do
                 xL' <- stepL xL a
@@ -184,9 +297,97 @@ instance Monad m => Applicative (FoldM m a) where
 instance (Monoid b, Monad m) => Monoid (FoldM m a b) where
     mempty = pure mempty
     {-# INLINABLE mempty #-}
+
     mappend = liftA2 mappend
     {-# INLINABLE mappend #-}
 
+instance (Monad m, Num b) => Num (FoldM m a b) where
+    fromInteger = pure . fromInteger
+    {-# INLINABLE fromInteger #-}
+
+    negate = fmap negate
+    {-# INLINABLE negate #-}
+
+    abs = fmap abs
+    {-# INLINABLE abs #-}
+
+    signum = fmap signum
+    {-# INLINABLE signum #-}
+
+    (+) = liftA2 (+)
+    {-# INLINABLE (+) #-}
+
+    (*) = liftA2 (*)
+    {-# INLINABLE (*) #-}
+
+    (-) = liftA2 (-)
+    {-# INLINABLE (-) #-}
+
+instance (Monad m, Fractional b) => Fractional (FoldM m a b) where
+    fromRational = pure . fromRational
+    {-# INLINABLE fromRational #-}
+
+    recip = fmap recip
+    {-# INLINABLE recip #-}
+
+    (/) = liftA2 (/)
+    {-# INLINABLE (/) #-}
+
+instance (Monad m, Floating b) => Floating (FoldM m a b) where
+    pi = pure pi
+    {-# INLINABLE pi #-}
+
+    exp = fmap exp
+    {-# INLINABLE exp #-}
+
+    sqrt = fmap sqrt
+    {-# INLINABLE sqrt #-}
+
+    log = fmap log
+    {-# INLINABLE log #-}
+
+    sin = fmap sin
+    {-# INLINABLE sin #-}
+
+    tan = fmap tan
+    {-# INLINABLE tan #-}
+
+    cos = fmap cos
+    {-# INLINABLE cos #-}
+
+    asin = fmap sin
+    {-# INLINABLE asin #-}
+
+    atan = fmap atan
+    {-# INLINABLE atan #-}
+
+    acos = fmap acos
+    {-# INLINABLE acos #-}
+
+    sinh = fmap sinh
+    {-# INLINABLE sinh #-}
+
+    tanh = fmap tanh
+    {-# INLINABLE tanh #-}
+
+    cosh = fmap cosh
+    {-# INLINABLE cosh #-}
+
+    asinh = fmap asinh
+    {-# INLINABLE asinh #-}
+
+    atanh = fmap atanh
+    {-# INLINABLE atanh #-}
+
+    acosh = fmap acosh
+    {-# INLINABLE acosh #-}
+
+    (**) = liftA2 (**)
+    {-# INLINABLE (**) #-}
+
+    logBase = liftA2 logBase
+    {-# INLINABLE logBase #-}
+
 -- | Apply a strict left 'Fold' to a 'Foldable' container
 fold :: Foldable f => Fold a b -> f a -> b
 fold (Fold step begin done) as = F.foldr cons done as begin
@@ -220,25 +421,21 @@ mconcat = Fold mappend mempty id
 
 -- | Convert a \"@foldMap@\" to a 'Fold'
 foldMap :: Monoid w => (a -> w) -> (w -> b) -> Fold a b
-foldMap to from = Fold (\x a -> mappend x (to a)) mempty from
+foldMap to = Fold (\x a -> mappend x (to a)) mempty
 {-# INLINABLE foldMap #-}
 
 {-| Get the first element of a container or return 'Nothing' if the container is
     empty
 -}
 head :: Fold a (Maybe a)
-head = Fold step Nothing' lazy
-  where
-    step x a = case x of
-        Nothing' -> Just' a
-        _        -> x
+head = _Fold1 const
 {-# INLINABLE head #-}
 
 {-| Get the last element of a container or return 'Nothing' if the container is
     empty
 -}
 last :: Fold a (Maybe a)
-last = Fold (\_ -> Just') Nothing' lazy
+last = _Fold1 (flip const)
 {-# INLINABLE last #-}
 
 {-| Get the last element of a container or return a default value if the container
@@ -299,20 +496,12 @@ product = Fold (*) 1 id
 
 -- | Computes the maximum element
 maximum :: Ord a => Fold a (Maybe a)
-maximum = Fold step Nothing' lazy
-  where
-    step x a = Just' (case x of
-        Nothing' -> a
-        Just' a' -> max a' a)
+maximum = _Fold1 max
 {-# INLINABLE maximum #-}
 
 -- | Computes the minimum element
 minimum :: Ord a => Fold a (Maybe a)
-minimum = Fold step Nothing' lazy
-  where
-    step x a = Just' (case x of
-        Nothing' -> a
-        Just' a' -> min a' a)
+minimum = _Fold1 min
 {-# INLINABLE minimum #-}
 
 {-| @(elem a)@ returns 'True' if the container has an element equal to @a@,
@@ -336,7 +525,7 @@ find :: (a -> Bool) -> Fold a (Maybe a)
 find predicate = Fold step Nothing' lazy
   where
     step x a = case x of
-        Nothing' -> if (predicate a) then Just' a else Nothing'
+        Nothing' -> if predicate a then Just' a else Nothing'
         _        -> x
 {-# INLINABLE find #-}
 
@@ -368,6 +557,25 @@ findIndex predicate = Fold step (Left' 0) hush
         _       -> x
 {-# INLINABLE findIndex #-}
 
+data Pair3 a b c = Pair3 !a !b !c
+
+-- | Pick a random element, using reservoir sampling
+random :: FoldM IO a (Maybe a)
+random = FoldM step begin done
+  where
+    begin = do
+        gen <- createSystemRandom
+        return $! Pair3 gen Nothing' (1 :: Int)
+
+    step (Pair3 gen Nothing'  _) a = return $! Pair3 gen (Just' a) 2
+    step (Pair3 gen (Just' a) m) b = do
+        n <- uniformR (1, m) gen
+        let c = if n == 1 then b else a
+        return $! Pair3 gen (Just' c) (m + 1)
+
+    done (Pair3 _ ma _) = return (lazy ma)
+{-# INLINABLE random #-}
+
 -- | Like 'length', except with a more general 'Num' return value
 genericLength :: Num b => Fold a b
 genericLength = Fold (\n _ -> n + 1) 0 id
@@ -378,7 +586,7 @@ genericIndex :: Integral i => i -> Fold a (Maybe a)
 genericIndex i = Fold step (Left' 0) done
   where
     step x a = case x of
-        Left'  j -> if (i == j) then Right' a else Left' (j + 1)
+        Left'  j -> if i == j then Right' a else Left' (j + 1)
         _        -> x
     done x = case x of
         Left'  _ -> Nothing
@@ -390,6 +598,11 @@ list :: Fold a [a]
 list = Fold (\x a -> x . (a:)) id ($ [])
 {-# INLINABLE list #-}
 
+-- | Fold all values into a list, in reverse order
+revList :: Fold a [a]
+revList = Fold (\x a -> a:x) [] id
+{-# INLINABLE revList #-}
+
 {-| /O(n log n)/.  Fold values into a list with duplicates removed, while
     preserving their first occurrences
 -}
@@ -415,7 +628,7 @@ eqNub = Fold step (Pair [] id) fin
 {-# INLINABLE eqNub #-}
 
 -- | Fold values into a set
-set :: (Ord a) => Fold a (Set.Set a)
+set :: Ord a => Fold a (Set.Set a)
 set = Fold (flip Set.insert) Set.empty id
 {-# INLINABLE set #-}
 
@@ -431,7 +644,7 @@ vector = FoldM step begin done
         return (Pair mv 0)
     step (Pair mv idx) a = do
         let len = M.length mv
-        mv' <- if (idx >= len)
+        mv' <- if idx >= len
             then M.unsafeGrow mv (min len maxChunkSize)
             else return mv
         M.unsafeWrite mv' idx a
@@ -504,10 +717,28 @@ simplify (FoldM step begin done) = Fold step' begin' done'
     done' x   = runIdentity (done x)
 {-# INLINABLE simplify #-}
 
+{-| @_Fold1 step@ returns a new 'Fold' using just a step function that has the
+same type for the accumulator and the element. The result type is the
+accumulator type wrapped in 'Maybe'. The initial accumulator is retrieved from
+the 'Foldable', the result is 'None' for empty containers.
+ -}
+_Fold1 :: (a -> a -> a) -> Fold a (Maybe a)
+_Fold1 step = Fold step_ Nothing' lazy
+  where
+    step_ mx a = Just' (case mx of
+        Nothing' -> a
+        Just' x -> step x a)
+
 {-| @(premap f folder)@ returns a new 'Fold' where f is applied at each step
 
 > fold (premap f folder) list = fold folder (map f list)
 
+>>> fold (premap Sum mconcat) [1..10]
+Sum {getSum = 55}
+
+>>> fold mconcat (map Sum [1..10])
+Sum {getSum = 55}
+
 > premap id = id
 >
 > premap (f . g) = premap g . premap f
@@ -541,47 +772,86 @@ premapM f (FoldM step begin done) = FoldM step' begin done
     step' x a = step x (f a)
 {-# INLINABLE premapM #-}
 
-type Traversal' a b = forall f . Applicative f => (b -> f b) -> a -> f a
+{-| A handler for the upstream input of a `Fold`
 
-{-| @(pretraverse t folder)@ traverses each incoming element using @Traversal'@
-    @t@ and folds every target of the @Traversal'@
-
-> pretraverse id = id
->
-> pretraverse (f . g) = pretraverse f . pretraverse g
-
-> pretraverse t (pure r) = pure r
->
-> pretraverse t (f <*> x) = pretraverse t f <*> pretraverse t x
+    Any lens, traversal, or prism will type-check as a `Handler`
 -}
-pretraverse :: Traversal' a b -> Fold b r -> Fold a r
-pretraverse k (Fold step begin done) = Fold step' begin done
+type Handler a b =
+    forall x . (b -> Constant (Endo x) b) -> a -> Constant (Endo x) a 
+
+{-| @(handles t folder)@ transforms the input of a `Fold` using a lens,
+    traversal, or prism:
+
+> handles _1       :: Fold a r -> Fold (a, b) r
+> handles _Left    :: Fold a r -> Fold (Either a b) r
+> handles traverse :: Traversable t => Fold a r -> Fold (t a) r
+
+>>> fold (handles traverse sum) [[1..5],[6..10]]
+55
+
+>>> fold (handles (traverse.traverse) sum) [[Nothing, Just 2, Just 7],[Just 13, Nothing, Just 20]]
+42
+
+>>> fold (handles (filtered even) sum) [1,3,5,7,21,21]
+42
+
+>>> fold (handles _2 mconcat) [(1,"Hello "),(2,"World"),(3,"!")]
+"Hello World!"
+
+> handles id = id
+>
+> handles (f . g) = handles f . handles g
+
+> handles t (pure r) = pure r
+>
+> handles t (f <*> x) = handles t f <*> handles t x
+-}
+handles :: Handler a b -> Fold b r -> Fold a r
+handles k (Fold step begin done) = Fold step' begin done
   where
     step' = flip (appEndo . getConstant . k (Constant . Endo . flip step))
-{-# INLINABLE pretraverse #-}
+{-# INLINABLE handles #-}
 
+{-|
+> instance Monad m => Monoid (EndoM m a) where
+>     mempty = EndoM return
+>     mappend (EndoM f) (EndoM g) = EndoM (f >=> g)
+-}
 newtype EndoM m a = EndoM { appEndoM :: a -> m a }
 
 instance Monad m => Monoid (EndoM m a) where
     mempty = EndoM return
-    mappend (EndoM f) (EndoM g) = EndoM (f <=< g)
+    mappend (EndoM f) (EndoM g) = EndoM (f >=> g)
 
-{-| @(pretraverseM t folder)@ traverses each incoming element using @Traversal'@
-    @t@ and folds every target of the @Traversal'@
+{-| A Handler for the upstream input of `FoldM`
 
-> pretraverseM id = id
->
-> pretraverseM (f . g) = pretraverseM f . pretraverseM g
-
-> pretraverseM t (pure r) = pure r
->
-> pretraverseM t (f <*> x) = pretraverseM t f <*> pretraverseM t x
+    Any lens, traversal, or prism will type-check as a `HandlerM`
 -}
-pretraverseM :: Monad m => Traversal' a b -> FoldM m b r -> FoldM m a r
-pretraverseM k (FoldM step begin done) = FoldM step' begin done
+type HandlerM m a b =
+    forall x . (b -> Constant (EndoM m x) b) -> a -> Constant (EndoM m x) a 
+
+{-| @(handlesM t folder)@ transforms the input of a `FoldM` using a lens,
+    traversal, or prism:
+
+> handlesM _1       :: FoldM m a r -> FoldM (a, b) r
+> handlesM _Left    :: FoldM m a r -> FoldM (Either a b) r
+> handlesM traverse :: Traversable t => FoldM m a r -> FoldM m (t a) r
+
+    `handlesM` obeys these laws:
+
+> handlesM id = id
+>
+> handlesM (f . g) = handlesM f . handlesM g
+
+> handlesM t (pure r) = pure r
+>
+> handlesM t (f <*> x) = handlesM t f <*> handlesM t x
+-}
+handlesM :: Monad m => HandlerM m a b -> FoldM m b r -> FoldM m a r
+handlesM k (FoldM step begin done) = FoldM step' begin done
   where
     step' = flip (appEndoM . getConstant . k (Constant . EndoM . flip step))
-{-# INLINABLE pretraverseM #-}
+{-# INLINABLE handlesM #-}
 
 {- $reexports
     @Control.Monad.Primitive@ re-exports the 'PrimMonad' type class
